from django.db import models


class RichTextField(models.TextField):

    def formfield(self, **kwargs):
        from django.conf import settings
        from django.utils.module_loading import import_string

        widget_path = getattr(
            settings,
            "PYMISSIVE_RICHTEXT_WIDGET",
            "django.forms.Textarea"
        )
        widget_class = import_string(widget_path)

        kwargs["widget"] = widget_class
        return super().formfield(**kwargs)


class JSONField(models.JSONField):

    def formfield(self, **kwargs):
        from django.conf import settings
        from django.utils.module_loading import import_string

        widget_path = getattr(settings, "PYMISSIVE_JSON_WIDGET", None)
        if widget_path:
            widget_class = import_string(widget_path)
            kwargs["widget"] = widget_class
        return super().formfield(**kwargs)


"""Admin for MissiveBilling model."""

from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from django_boosted import AdminBoostModel
from ..models.billing import MissiveBilling


class MissiveBillingInline(admin.TabularInline):
    """Inline for billing records on missive."""

    model = MissiveBilling
    extra = 0
    readonly_fields = ["created_at"]
    fields = ["recipient", "billing_amount", "estimate_amount", "currency", "is_billed", "invoice", "created_at"]
    raw_id_fields = ["recipient"]
    can_delete = True
    show_change_link = True
    readonly_fields = [
        "missive",
        "recipient",
        "billing_amount",
        "estimate_amount",
        "currency",
        "is_billed",
        "invoice",
        "created_at",
    ]

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


@admin.register(MissiveBilling)
class MissiveBillingAdmin(AdminBoostModel):
    """Admin for billing records."""

    list_display = [
        "missive",
        "recipient",
        "billing_amount",
        "estimate_amount",
        "is_billed",
        "created_at",
    ]
    list_filter = [
        "missive__provider",
        "missive__missive_type",
        "is_billed"
    ]
    search_fields = [
        "missive__external_id",
        "missive__subject",
        "recipient__name",
        "recipient__email",
        "recipient__phone",
    ]
    raw_id_fields = ["missive", "recipient"]
    readonly_fields = ["created_at", "updated_at"]
    actions = ["set_billed"]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "missive",
                    "recipient",
                    "billing_amount",
                    "estimate_amount",
                    "currency",
                    "invoice",
                    "is_billed",
                )
            },
        ),
    ]
    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(_("Comment/Timestamps"), ["comment", "created_at", "updated_at"])

    @admin.action(description=_("Mark as billed"))
    def set_billed(self, request, queryset):
        updated = queryset.filter(billing_amount__gt=0).update(is_billed=True)
        self.message_user(request, _(f"{updated} record(s) marked as billed."))



from celery import shared_task
from .base import BaseCampaignBackend
from .campaign import run_campaign


@shared_task
def celery_campaign_task(campaign_id):
    run_campaign(campaign_id)


class CeleryBackend(BaseCampaignBackend):
    def delay(self, campaign_id: int):
        celery_campaign_task.delay(campaign_id)



"""MissiveRelatedObject model for linking missives to other models."""

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _

from .base import CommentTimestampedModel
from ..managers.related_object import MissiveRelatedObjectManager, CampaignRelatedObjectManager


class BaseRelatedObject(CommentTimestampedModel):
    """Model to link a missive or campaign to any other Django model."""

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        verbose_name=_("Content Type"),
        help_text=_("Type of the related object"),
    )
    object_id = models.PositiveIntegerField(
        verbose_name=_("Object ID"),
        help_text=_("ID of the related object"),
    )
    content_object = GenericForeignKey("content_type", "object_id")
    object_str = models.CharField(
        editable=False,
        max_length=500,
        verbose_name=_("Object String Representation"),
        help_text=_(
            "String representation of the related object (saved for reference if object is deleted)"
        ),
    )
    objects = MissiveRelatedObjectManager()

    class Meta:
        abstract = True
        verbose_name = _("Related Object")
        verbose_name_plural = _("Related Objects")
        ordering = ["-created_at"]

    def save(self, *args, **kwargs):
        """Save the object string representation before saving."""
        if self.content_object:
            try:
                self.object_str = str(self.content_object)
            except Exception:
                pass
        super().save(*args, **kwargs)


class MissiveRelatedObject(BaseRelatedObject):
    """Model to link a missive to any other Django model."""
    missive = models.ForeignKey(
        "django_pymissive.Missive",
        on_delete=models.CASCADE,
        related_name="to_missiverelatedobject",
        verbose_name=_("Missive"),
        help_text=_("Missive to which this object is related"),
    )

    objects = MissiveRelatedObjectManager()

    class Meta:
        verbose_name = _("Missive Related Object")
        verbose_name_plural = _("Missive Related Objects")
        ordering = ["-created_at"]

    def __str__(self):
        if self.content_object:
            return f"{self.missive} -> {self.content_object}"
        elif self.object_str:
            return f"{self.missive} -> {self.object_str} (deleted)"
        return f"{self.missive} -> {self.content_type} #{self.object_id}"


class CampaignRelatedObject(BaseRelatedObject):
    """Model to link a campaign to any other Django model."""
    campaign = models.ForeignKey(
        "django_pymissive.MissiveCampaign",
        on_delete=models.CASCADE,
        related_name="to_campaignrelatedobject",
        verbose_name=_("Campaign"),
        help_text=_("Campaign to which this object is related"),
    )

    objects = CampaignRelatedObjectManager()

    class Meta:
        verbose_name = _("Campaign Related Object")
        verbose_name_plural = _("Campaign Related Objects")
        ordering = ["-created_at"]
    
    def __str__(self):
        if self.content_object:
            return f"{self.campaign} -> {self.content_object}"
        elif self.object_str:
            return f"{self.campaign} -> {self.object_str} (deleted)"
        return f"{self.campaign} -> {self.content_type} #{self.object_id}"



from django_providerkit.models.service import ProviderServiceModelBase
from django.utils.translation import gettext_lazy as _
from django_providerkit.managers.service import ProviderServiceManager

class MissiveServiceModel(ProviderServiceModelBase):
    """Virtual model for missive services."""

    objects = ProviderServiceManager(package_name='pymissive')

    class Meta:
        managed = False
        verbose_name = _('Provider Service')
        verbose_name_plural = _('Provider Services')


"""MissiveEvent model for tracking missive events."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base import CommentTimestampedModel
from .choices import MissiveEventType
from ..managers.event import MissiveEventManager
from django.utils import timezone
from ..fields import JSONField


class MissiveEvent(CommentTimestampedModel):
    """Event tracking for missives (status changes, webhooks, etc.)."""

    missive = models.ForeignKey(
        "django_pymissive.Missive",
        on_delete=models.CASCADE,
        related_name="to_missiveevent",
        verbose_name=_("Missive"),
        help_text=_("Missive associated with this event"),
        blank=True,
        null=True,
        editable=False,
    )

    recipient = models.ForeignKey(
        "django_pymissive.MissiveRecipient",
        on_delete=models.CASCADE,
        related_name="to_recipientevent",
        verbose_name=_("Recipient"),
        help_text=_("Recipient associated with this event"),
        blank=True,
        null=True,
        editable=False,
    )

    event = models.CharField(
        max_length=50,
        choices=MissiveEventType.choices,
        null=True,
        blank=True,
        verbose_name=_("Event"),
        help_text=_("Event type (sent, delivered, read, failed, etc.)"),
    )

    reason = models.TextField(
        blank=True,
        verbose_name=_("Reason"),
        help_text=_("Reason or details about this event (provider-specific)"),
    )

    metadata = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Metadata"),
        help_text=_("Additional metadata as JSON"),
    )

    trace = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Trace"),
        help_text=_("Raw trace data (webhook payload, API response, etc.)"),
    )

    client_initiated = models.BooleanField(
        default=False,
        verbose_name=_("Client Initiated"),
        help_text=_(
            "True if the event was initiated by django_pymissive (send, resend, cancel, batch, etc.), "
            "False if received from a provider webhook."
        ),
    )

    occurred_at = models.DateTimeField(
        verbose_name=_("Occurred At"),
        help_text=_("When this event occurred"),
    )

    objects = MissiveEventManager()

    class Meta:
        verbose_name = _("Event")
        verbose_name_plural = _("Events")
        ordering = ["-occurred_at",]

    def get_reason(self):
        """Return human-readable reason for event from config, or empty string."""
        return MissiveEventType.get_description(self.event)

    def save(self, *args, **kwargs):
        if not self.occurred_at:
            self.occurred_at = timezone.now()
        if not self.reason and self.event:
            self.reason = self.get_reason()
        super().save(*args, **kwargs)

    def __str__(self):
        missive_ref = self.missive_id or "?"
        return f"{missive_ref} - {self.event} ({self.occurred_at})"

    def can_replay(self):
        """Return True if the event can be replayed."""
        return (self.missive_id and not self.client_initiated)

    def replay(self):
        """Replay the event. Uses trace["raw"] (normalized event) stored at creation."""
        if not self.missive_id:
            raise ValueError("Cannot replay event without associated missive")



        event = self.trace.get("raw") or self.trace
        if not isinstance(event, dict):
            raise ValueError("No replayable event in trace")
        event = dict(event)
        event["pk"] = self.pk
        self.missive.handle_events([event])



"""Admin for provider model."""

from django.contrib import admin

from django_providerkit.admin.provider import BaseProviderAdmin

from ..models.provider import MissiveProviderModel


@admin.register(MissiveProviderModel)
class ProviderAdmin(BaseProviderAdmin):
    """Admin for missive providers."""

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



"""
Sync events for missives created since a given date.

Deletes events that are not client_initiated (i.e. provider webhook events),
re-fetches events from the provider when available, and updates status.

Usage:
  ./manage.py sync_events --date 2026-03-01
  ./manage.py sync_events -d 2026-01-15
  ./manage.py sync_events -d 2026-01-15 --noevent
"""

from datetime import datetime

from django.db.models import Count
from django.core.management.base import BaseCommand

from ...models.missive import Missive


class Command(BaseCommand):
    help = (
        "Sync events for missives created since a given date: "
        "delete non-client-initiated events, re-fetch from provider, update status."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--date",
            required=True,
            help="Minimum creation date (YYYY-MM-DD). Missives created on or after this date are processed.",
        )
        parser.add_argument(
            "-n",
            "--noevent",
            action="store_true",
            help="Only process missives that have 0 or 1 event.",
        )

    def handle(self, *args, **options):
        date_str = options["date"]
        try:
            since = datetime.strptime(date_str, "%Y-%m-%d").date()
        except ValueError:
            self.stderr.write(
                self.style.ERROR(f"Invalid date format: {date_str}. Use YYYY-MM-DD.")
            )
            return

        missives = (
            Missive.objects.filter(created_at__date__gte=since, external_id__isnull=False)
            .exclude(external_id="")
            .order_by("created_at")
        )
        if options.get("noevent"):
            missives = missives.annotate(event_count=Count("to_missiveevent")).filter(
                event_count__lte=1
            )
        total = missives.count()
        synced = 0
        errors = 0

        for missive in missives:
            try:
                deleted, _ = missive.to_missiveevent.filter(
                    client_initiated=False
                ).delete()

                if missive.has_service("retrieve"):
                    missive.retrieve_missive()

                missive.set_status()
                for recipient in missive.to_missiverecipient.all():
                    recipient.set_status()

                synced += 1
                if options.get("verbosity", 1) >= 2:
                    self.stdout.write(
                        f"Synced events for missive {missive.id} (deleted {deleted} events)"
                    )
            except Exception as e:
                errors += 1
                self.stderr.write(
                    self.style.ERROR(f"Error syncing missive {missive.id}: {e}")
                )

        self.stdout.write(
            self.style.SUCCESS(
                f"Done: {synced} synced, {errors} errors (total: {total})"
            )
        )



import django_rq
from .base import BaseCampaignBackend
from .campaign import run_campaign


class RQBackend(BaseCampaignBackend):
    def delay(self, campaign_id: int):
        queue = django_rq.get_queue("default")
        queue.enqueue(run_campaign, campaign_id)



from threading import Thread
from .base import BaseCampaignBackend
from .campaign import run_campaign


class ThreadBackend(BaseCampaignBackend):
    def delay(self, campaign_id: int):
        Thread(target=run_campaign, args=(campaign_id,), daemon=True).start()



"""Manager for MissiveCampaign model."""

from django.db import models
from django.db.models import Case, F, Q, Value, When
from django.db.models.expressions import Subquery, OuterRef
from django.db.models.functions import Coalesce
from pymissive.config import MISSIVE_TYPES

from ..models.choices import MissiveStatus


class MissiveCampaignManager(models.Manager):
    """Manager for MissiveCampaign with annotated counts."""

    def last_scheduled_subquery(self, field: str = "event"):
        from ..models.campaign import MissiveScheduledCampaign

        return Subquery(
            MissiveScheduledCampaign.objects.filter(
                campaign=OuterRef("pk"),
            )
            .order_by(f"-{field}", "-id")
            .values(field)[:1],
            output_field=models.CharField(),
        )

    def pct_expr(self, cnt):
        return Coalesce(
            Case(
                When(count_recipient=0, then=Value(0.0)),
                default=cnt * 100.0 / F("count_recipient"),
                output_field=models.FloatField(),
            ),
            Value(0.0),
        )

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.annotate(
            last_send_date=self.last_scheduled_subquery("send_date"),
            last_ended_at=self.last_scheduled_subquery("ended_at"),
            count_missive=models.Count("to_missive", distinct=True),
            count_recipient=models.Count("to_missive__to_missiverecipient", distinct=True),
            count_event=models.Count("to_missive__to_missiveevent", distinct=True),
            count_related_object=models.Count("to_campaignrelatedobject", distinct=True),
            count_attachment=models.Count("to_campaigndocument", distinct=True),
            **{
                f"count_missive_{status.value.lower()}": models.Count(
                    "to_missive",
                    filter=Q(to_missive__status=status),
                    distinct=True,
                )
                for status in MissiveStatus
            },
            **{
                f"count_type_{type_key}": models.Count(
                    "to_missive",
                    filter=Q(to_missive__missive_type=type_key),
                    distinct=True,
                )
                for type_key in MISSIVE_TYPES
            },
            **{
                f"count_recipient_{status.value.lower()}": models.Count(
                    "to_missive__to_missiverecipient",
                    distinct=True,
                    filter=Q(to_missive__to_missiverecipient__status=status),
                )
                for status in MissiveStatus
            },
        )
        qs = qs.annotate(
            **{
                f"pct_recipient_{status.value.lower()}": self.pct_expr(F(f"count_recipient_{status.value.lower()}"))
                for status in MissiveStatus
            },
        )
        return qs



"""Manager for missive providers."""

from django_providerkit.managers import BaseProviderManager


class ProviderManager(BaseProviderManager):
    """Manager for missive providers."""

    package_name = "pymissive"



from django.db import models


class MissiveRelatedObjectManager(models.Manager):
    """Manager for the MissiveRelatedObject model."""

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.select_related("missive")
        return qs


class CampaignRelatedObjectManager(models.Manager):
    """Manager for the CampaignRelatedObject model."""

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.select_related("campaign")
        return qs


"""
Sync billings for missives created since a given date.

Deletes existing billings and recreates them from the provider.

Usage:
  ./manage.py sync_billings --date 2026-03-01
  ./manage.py sync_billings -d 2026-01-15
"""

from datetime import datetime

from django.core.management.base import BaseCommand

from ...models.missive import Missive


class Command(BaseCommand):
    help = (
        "Sync billings for missives created since a given date: "
        "delete existing billings and recreate them from the provider."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--date",
            required=True,
            help="Minimum creation date (YYYY-MM-DD). Missives created on or after this date are processed.",
        )

    def handle(self, *args, **options):
        date_str = options["date"]
        try:
            since = datetime.strptime(date_str, "%Y-%m-%d").date()
        except ValueError:
            self.stderr.write(
                self.style.ERROR(f"Invalid date format: {date_str}. Use YYYY-MM-DD.")
            )
            return

        missives = (
            Missive.objects.filter(created_at__date__gte=since)
            .filter(external_id__isnull=False)
            .exclude(external_id="")
            .order_by("created_at")
        )
        total = missives.count()
        synced = 0
        skipped = 0
        errors = 0

        for missive in missives:
            if not missive.can_billings():
                skipped += 1
                if options.get("verbosity", 1) >= 2:
                    self.stdout.write(
                        f"Skipped {missive.id} (no get_billings or no external_id)"
                    )
                continue

            try:
                missive.to_missivebilling.all().delete()
                missive.get_billings()
                synced += 1
                if options.get("verbosity", 1) >= 2:
                    self.stdout.write(
                        self.style.SUCCESS(f"Synced billings for missive {missive.id}")
                    )
            except Exception as e:
                errors += 1
                self.stderr.write(
                    self.style.ERROR(f"Error syncing missive {missive.id}: {e}")
                )

        self.stdout.write(
            self.style.SUCCESS(
                f"Done: {synced} synced, {skipped} skipped, {errors} errors (total: {total})"
            )
        )



"""Default PDF generator for django-pymissive.

Compiles the missive body template and converts it to PDF using weasyprint.
Override by setting PYMISSIVE_PDF_GENERATOR in your Django settings.
"""

from __future__ import annotations

from typing import Any

from django.template import Context, Template

HTML_WRAPPER = """<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8">
<style>
@page {{
    size: A4;
    margin: 20mm;
}}
body {{
    font-family: serif;
    font-size: 12pt;
    line-height: 1.5;
    color: #000;
}}
</style>
{extra_css}
</head>
<body>
{body}
</body>
</html>"""


def _compile_body(missive) -> str:
    """Compile first_document (campaign) or body_html when no campaign."""
    context = missive.missive_context()
    tpl = missive.get_locally_or_campaign_value("first_document", missive.body_html)
    if not tpl:
        tpl = missive.get_locally_or_campaign_value("body_html", "")
    return Template(str(tpl or "")).render(Context(context))


def body_to_pdf(missive, **kwargs: Any) -> bytes:
    """Compile the body and convert it to a PDF (bytes).

    Requires ``weasyprint`` to be installed::

        pip install django-pymissive[pdf]

    Accepted kwargs:
        extra_css (str): additional ``<style>`` block injected in ``<head>``.
    """
    try:
        from weasyprint import HTML  # type: ignore[import-untyped]
    except ImportError as exc:
        raise ImportError(
            "weasyprint is required for PDF generation. "
            "Install it with: pip install django-pymissive[pdf]"
        ) from exc

    compiled_body = _compile_body(missive)
    extra_css = kwargs.get("extra_css", "")
    if extra_css:
        extra_css = f"<style>{extra_css}</style>"

    html_string = HTML_WRAPPER.format(body=compiled_body, extra_css=extra_css)
    return HTML(string=html_string).write_pdf()



"""Webhook view for receiving provider events."""

import logging

from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import DetailView

from ..events import handle_events
from ..models.provider import MissiveProviderModel

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class WebhookView(DetailView):
    """Webhook view based on provider model."""

    model = MissiveProviderModel
    slug_field = "name"
    slug_url_kwarg = "provider"

    def post(self, request, *args, **kwargs):
        """Handle webhook POST request."""
        return self.handle_webhook(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        """Handle webhook GET request."""
        return self.handle_webhook(request, *args, **kwargs)

    def handle_webhook(self, request, *args, **kwargs):
        provider = self.get_object()
        missive_type = kwargs.get("missive_type")
        try:
            handle_events(request.body, provider=provider, missive_type=missive_type)
        except Exception as e:
            logger.error(f"Error handling webhook: {e}", exc_info=True)
        return HttpResponse(status=200)



"""Main Missive model for multi-channel sending."""

import uuid
from django.db import transaction
from django.utils import timezone
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.template import Context, Template
from django_providerkit import ProviderField
from django.utils.safestring import mark_safe
from django.urls import reverse
from django_geoaddress.fields import GeoaddressField
from phonenumber_field.modelfields import PhoneNumberField
from django.conf import settings
from django.utils.module_loading import import_string
from .choices import (
    AcknowledgementLevel,
    MissiveSupport,
    MissiveEventType,
    MissivePriority,
    MissiveStatus,
    status_from_event_counts,
    MissiveType,
    get_missive_support_from_type,
    MissiveRecipientType,
    MissiveAttachmentType,
    MissiveThreadType,
    MissiveDeliveryMode,
)
from ..managers import (
    MissiveManager,
    MissiveMessageManager,
    MissiveHistoryManager,
)
from ..models.base import CommentTimestampedModel
from ..fields import RichTextField, JSONField
from ..dispatch_signals import (
    missive_post_duplicate,
    missive_post_send,
    missive_pre_duplicate,
    missive_pre_send,
)
from ..utils import get_base_url, build_webhook_url, get_default_domain, get_default_scheme
from django.core import signing
from django.core.files.base import ContentFile


SEPARATOR = "\n--------------------------------\n"
ATTACHMENT_ICON = "&#128196;"
ATTACHMENT_STYLE = "text-decoration: none; font-size: 14px;"
ATTACHMENT_TPL_HTML = """<div>
    <a href='{url}' target='_blank' rel='noopener' style='{style}'>
        {icon}&nbsp;{name}
    </a>
</div>"""

PREVIEW_ICON = "&#127760;"
PREVIEW_STYLE = "text-decoration: none; font-size: 14px;"
PREVIEW_TPL_HTML = """<a href='{url}' target='_blank' rel='noopener' style='{style}'>
    {icon}&nbsp;{text}
</a>"""


class Missive(CommentTimestampedModel):
    """Multi-channel missive model (email, SMS, address/LRE, application, etc.)."""
    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        verbose_name=_("ID"),
    )
    campaign = models.ForeignKey(
        "django_pymissive.MissiveCampaign",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="to_missive",
        verbose_name=_("Campaign"),
        help_text=_("Optional campaign this missive belongs to"),
    )
    thread_id = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        verbose_name=_("Thread"),
        help_text=_("Thread ID for the missive"),
        db_index=True,
    )
    thread_type = models.CharField(
        max_length=50,
        choices=MissiveThreadType.choices,
        default=MissiveThreadType.MISSIVE,
        verbose_name=_("Thread Type"),
        help_text=_("Type of thread (missive, message, history)"),
    )
    provider = ProviderField(
        package_name="pymissive",
        blank=True,
        verbose_name=_("Provider"),
        help_text=_("Provider used to send this missive"),
    )
    status = models.CharField(
        max_length=20,
        choices=MissiveStatus.choices,
        default=MissiveStatus.DRAFT,
        verbose_name=_("Status"),
        help_text=_("Current status of the missive"),
    )
    missive_support = models.CharField(
        max_length=50,
        choices=MissiveSupport.choices,
        verbose_name=_("Missive Support"),
        help_text=_("Support for the missive (email, phone, address, application)"),
        editable=False,
    )
    brand_name = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        verbose_name=_("Brand Name"),
        help_text=_("Brand name used to send this missive"),
    )
    missive_type = models.CharField(
        max_length=50,
        choices=MissiveType.choices,
        verbose_name=_("Missive Type"),
        help_text=_("Type of missive (email, sms, lre, ere, etc.)"),
    )
    acknowledgement = models.CharField(
        max_length=50,
        choices=AcknowledgementLevel.choices,
        blank=True,
        null=True,
        verbose_name=_("Acknowledgement Level"),
        help_text=_("Desired acknowledgement level for delivery proof"),
    )
    delivery_mode = models.CharField(
        max_length=50,
        choices=MissiveDeliveryMode.choices,
        blank=True,
        null=True,
        verbose_name=_("Delivery Mode"),
        help_text=_("Delivery mode (economic, normal, premium, express)"),
    )
    priority = models.CharField(
        max_length=20,
        choices=MissivePriority.choices,
        blank=True,
        null=True,
        verbose_name=_("Priority"),
        help_text=_("Priority level"),
    )
    subject = models.CharField(
        max_length=500,
        verbose_name=_("Subject"),
        help_text=_("Subject line (for email, SMS, etc.)"),
        blank=True,
        null=True,
    )

    body_html = RichTextField(
        blank=True,
        null=True,
        verbose_name=_("Body HTML"),
        help_text=_("HTML message body (email) or rich content (LRE)"),
    )
    body_text = models.TextField(
        blank=True,
        null=True,
        verbose_name=_("Body Text"),
        help_text=_("Plain text version of the message"),
    )

    # Sender
    sender_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name=_("Sender name"),
        help_text=_("Display name of the sender"),
    )
    sender_email = models.EmailField(
        blank=True,
        null=True,
        verbose_name=_("Sender email"),
        help_text=_("Email address of the sender"),
    )
    sender_phone = PhoneNumberField(
        blank=True,
        null=True,
        verbose_name=_("Sender phone"),
        help_text=_("Phone number of the sender (used for SMS)"),
    )
    sender_address = GeoaddressField(
        blank=True,
        null=True,
        verbose_name=_("Sender address"),
        help_text=_("Postal address of the sender"),
    )

    # Reply-To (email only)
    reply_to_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name=_("Reply-To name"),
        help_text=_("Display name for reply-to address"),
    )
    reply_to_email = models.EmailField(
        blank=True,
        null=True,
        verbose_name=_("Reply-To email"),
        help_text=_("Email address for replies"),
    )
    reply_to_address = GeoaddressField(
        max_length=512,
        blank=True,
        null=True,
        verbose_name=_("Reply-To address"),
        help_text=_("Postal address for replies"),
    )
    additional_context = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Additional context"),
        help_text=_("Additional context as JSON"),
    )
    external_id = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        editable=False,
        verbose_name=_("External ID"),
        help_text=_("External identifier from the provider"),
    )
    metadata = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Metadata"),
        help_text=_("Additional metadata as JSON"),
    )
    additional_config = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Additional configuration"),
        help_text=_("Additional configuration as JSON"),
    )
    webhook_url = models.URLField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name=_("Webhook URL"),
        help_text=_("Webhook URL for the missive"),
    )
    message_by = models.ForeignKey(
        "django_pymissive.MissiveRecipient",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="to_missivemessageby",
        verbose_name=_("Reply by"),
        help_text=_("Recipient who sent this reply (for inbound exchanges)"),
    )

    objects = MissiveManager()

    class Meta:
        verbose_name = _("Missive")
        verbose_name_plural = _("Missives")
        ordering = ["-created_at"]

    def __str__(self):
        recipient = self.first_recipient or _("Unknown")
        return f"{self.missive_type} - {recipient} ({self.status})"

    def _ensure_default_provider(self):
        """Set provider from MissiveConfig if empty."""
        if self.provider or not self.missive_type:
            return
        from .config import MissiveConfig
        config = MissiveConfig.objects.filter(missive_type=self.missive_type).first()
        if config and config.default_provider:
            self.provider = config.default_provider

    def _ensure_missive_defaults(self):
        """Apply default values for support and delivery settings."""
        if self.can_be_modified:
            support = get_missive_support_from_type(self.missive_type)
            if support:
                self.missive_support = support
        # Leave empty when campaign assigned; get_serialized_data resolves via locally_or_campaign
        # Use campaign_id to avoid FK resolution issues when creating via API/viewset
        has_campaign = bool(self.campaign_id)
        if not self.acknowledgement and not has_campaign:
            self.acknowledgement = AcknowledgementLevel.BASIC_DELIVERY
        if not self.delivery_mode and not has_campaign:
            self.delivery_mode = MissiveDeliveryMode.NORMAL
        if not self.priority and not has_campaign:
            self.priority = MissivePriority.NORMAL

    def save(self, *args, **kwargs):
        """Save the missive with auto-filled defaults (provider, support, acknowledgement, etc.)."""
        self._ensure_default_provider()
        self._ensure_missive_defaults()
        super().save(*args, **kwargs)

    def has_service(self, service):
        service_name = f"{service}_{self.missive_type}".lower()
        if self.provider:
            return hasattr(self.provider._provider, service_name)

    @property
    def token_missive(self):
        data = {"id": str(self.id)}
        return signing.dumps(data)

    @property
    def can_be_modified(self):
        return not self.external_id

    @property
    def last_event_display(self):
        return dict(MissiveEventType.choices).get(self.last_event, self.last_event)

    # Campaign suffix by support: address->lre, email->email, phone->phone
    _SUPPORT_TO_CAMPAIGN_SUFFIX = {"address": "lre", "email": "email", "phone": "phone"}

    @classmethod
    def get_campaign_sourced_fields(cls, support):
        """Return (missive_attr, lookup_field) for fields that can come from campaign."""
        support = (support or "").lower()
        fields = [
            ("subject", "subject"),
            ("body_html", "body_html"),
            ("body_text", "body_text"),
            ("acknowledgement", "acknowledgement"),
            ("delivery_mode", "delivery_mode"),
            ("priority", "priority"),
            ("sender_name", f"sender_{support}_name"),
            (f"sender_{support}", f"sender_{support}"),
            ("reply_to_name", f"reply_to_{support}_name"),
            (f"reply_to_{support}", f"reply_to_{support}"),
        ]
        if support == "phone":
            fields.append(("body_text", "body_sms"))
        elif support == "address":
            fields.append(("body_html", "first_document"))
        return fields

    @property
    def campaign_sourced_field_names(self):
        """Field names that can be sourced from campaign (for set_locally_ifnull / clear)."""
        fields = self.get_campaign_sourced_fields(self.missive_support)
        names = []
        for missive_attr, _ in fields:
            if missive_attr not in names and hasattr(self, missive_attr):
                names.append(missive_attr)
        return names + (["additional_context"] if self.campaign_id else [])

    def get_locally_or_campaign_value(self, field, fallback=None):
        """Return value from self, else from campaign (field or field_{suffix})."""
        locally = getattr(self, field, fallback)
        if locally:
            return locally
        if not self.campaign:
            return fallback
        campaign_val = getattr(self.campaign, field, None)
        if campaign_val:
            return campaign_val
        support = (self.missive_support or "").lower()
        if support:
            suffix = self._SUPPORT_TO_CAMPAIGN_SUFFIX.get(support, support)
            campaign_val = getattr(self.campaign, f"{field}_{suffix}", None)
            if campaign_val:
                return campaign_val
        return fallback

    def set_locally_ifnull(self):
        """Copy campaign values to local fields when null. Preserves content at send time
        so that if the campaign changes later, already-sent missives keep their content."""
        if not self.campaign_id:
            return
        support = (self.missive_support or "").lower()
        fields = self.get_campaign_sourced_fields(support)

        updates = []
        for missive_attr, lookup_field in fields:
            if not hasattr(self, missive_attr):
                continue
            local = getattr(self, missive_attr, None)
            val = self.get_locally_or_campaign_value(lookup_field, local)
            if not local and val:
                setattr(self, missive_attr, val)
                updates.append(missive_attr)

        if self.campaign.additional_context and not (self.additional_context or {}):
            self.additional_context = dict(self.campaign.additional_context)
            updates.append("additional_context")

        if updates:
            self.save(update_fields=updates)

    def clear_campaign_sourced_fields(self, missive):
        """Clear campaign-sourced fields on missive so they will be re-filled from campaign at send."""
        fields_to_clear = missive.campaign_sourced_field_names
        if not fields_to_clear:
            return
        for attr in fields_to_clear:
            if not hasattr(missive, attr):
                continue
            empty = {} if attr == "additional_context" else None
            setattr(missive, attr, empty)
        missive.save(update_fields=fields_to_clear)

    @property
    def sender(self):
        return self.get_sender()

    @property
    def reply_to(self):
        return self.get_reply_to()

    def get_sender(self):
        support = self.missive_support.lower()
        name = self.get_locally_or_campaign_value(f"sender_{support}_name", self.sender_name)
        sender = self.get_locally_or_campaign_value(f"sender_{support}", getattr(self, f"sender_{support}", None))
        sender = dict(sender) if support == "address" else str(sender) if sender else ""
        return {
            "name": name or "",
            support: sender,
        }

    def get_reply_to(self):
        """Return reply_to dict for provider (email only)."""
        support = self.missive_support.lower()
        name = self.get_locally_or_campaign_value(f"reply_to_{support}_name", self.reply_to_name)
        reply_to = self.get_locally_or_campaign_value(f"reply_to_{support}", getattr(self, f"reply_to_{support}", None))
        if reply_to:
            return {
                "name": name or "",
                support: str(reply_to),
            }
        return None

    def get_acknowledgement(self):
        return self.get_locally_or_campaign_value(
            "acknowledgement", fallback=AcknowledgementLevel.BASIC_DELIVERY
        )

    def get_delivery_mode(self):
        return self.get_locally_or_campaign_value(
            "delivery_mode", fallback=MissiveDeliveryMode.NORMAL
        )

    def get_priority(self):
        return self.get_locally_or_campaign_value(
            "priority", fallback=MissivePriority.NORMAL
        )

    def get_webhook_url(self):
        scheme = get_default_scheme()
        domain = get_default_domain()
        base = f"{scheme}://{(domain or '').strip().lstrip('/')}"
        return build_webhook_url(base, self.provider._provider.name, self.missive_type)

    def is_serializable_field(self, field):
        return (not field.is_relation
                and not field.many_to_many
                and not field.name.startswith("_"))

    def get_serialized_data(self, attachments=True):
        """Serialize missive data to a dictionary for provider calls."""

        missive_data = {}
        for field in self._meta.get_fields():
            if self.is_serializable_field(field):
                if hasattr(self, f"get_{field.name}"):
                    missive_data[field.name] = getattr(self, f"get_{field.name}")()
                elif hasattr(self, f"{field.name}_compiled"):
                    missive_data[field.name] = getattr(self, f"{field.name}_compiled")
                else:
                    missive_data[field.name] = getattr(self, field.name)
        missive_data["recipients"] = [
            recipient.get_serialized_data() for recipient in self.recipients
        ]
        if self.cc:
            missive_data["cc"] = [
                recipient.get_serialized_data() for recipient in self.cc
            ]
        if self.bcc:
            missive_data["bcc"] = [
                recipient.get_serialized_data() for recipient in self.bcc
            ]
        missive_data["sender"] = self.get_sender()
        missive_data["reply_to"] = self.get_reply_to()
        if attachments:
            missive_data["attachments"] = self.get_serialized_attachments(linked=False)
        missive_data["webhook_url"] = self.get_webhook_url()
        missive_data.update(self.additional_config)
        return missive_data

    def call_provider_service(self, service: str, **kwargs):
        """Call a provider service."""
        service_name = f"{service}_{self.missive_type}".lower()
        return self.provider.call_service(service_name,  **kwargs)

    #########################################################
    # Check methods
    #########################################################

    def can_send(self):
        if self.has_service("send") and (not self.external_id or self.status == MissiveStatus.DRAFT):
            service_method = f"check_{self.missive_type}"
            return getattr(self, service_method)() if hasattr(self, service_method) else True
        return False

    def can_resend(self):
        if self.has_service("send"):
            service_method = f"check_{self.missive_type}"
            return getattr(self, service_method)() if hasattr(self, service_method) else True
        return False

    def check_recipients(self):
        return self.recipients.filter(recipient_type=MissiveRecipientType.RECIPIENT).exists()

    def check_email(self):
        body_html = self.get_locally_or_campaign_value("body_html")
        body_text = self.get_locally_or_campaign_value("body_text")
        subject = self.get_locally_or_campaign_value("subject")
        body = body_html or body_text
        return self.check_recipients() and bool(body and body.strip()) and bool(subject and subject.strip())

    def check_sms(self):
        body = self.get_locally_or_campaign_value("body_sms", self.body_text)
        return self.check_recipients() and bool(body and body.strip())

    def check_lre(self):
        body = self.get_locally_or_campaign_value("first_document", self.body_html)
        return self.check_recipients() and bool(body and body.strip())

    @property
    def show_preview_browser(self):
        """Show the preview browser."""
        url = reverse("django_pymissive:preview", args=["missive", self.pk])
        url = self.base_url + url
        data = {
            "url": url,
            "icon": PREVIEW_ICON,
            "text": _("Preview in browser"),
            "style": PREVIEW_STYLE,
        }
        return mark_safe(PREVIEW_TPL_HTML.format(**data))  # nosec B703 B308

    @property
    def show_preview_browser_text(self):
        """Show the preview browser text."""
        url = reverse("django_pymissive:preview", args=["missive", self.pk])
        url = self.base_url + url
        return f"- {_('Preview in browser')}:{SEPARATOR}{url}\n"

    def missive_context(self):
        """Get the context of the missive."""
        context = getattr(self.campaign, "additional_context", {})
        context.update(self.additional_context or {})
        context.update({
            "show_preview_browser": self.show_preview_browser,
            "show_preview_browser_text": self.show_preview_browser_text,
            "show_attachments_linked": self.show_attachments_linked,
            "show_attachments_linked_text": self.show_attachments_linked_text,
        })
        return context

    def body_to_pdf(self):
        """Convert the body to PDF."""
        pdg_generator = getattr(settings, "MISSIVEPDF_GENERATOR", "django_pymissive.pdf.body_to_pdf")
        pdf = import_string(pdg_generator)(self)
        return pdf

    def generate_first_document(self):
        """Generate first document PDF from first_document/body_html and save as attachment."""
        from ..models.attachment import MissiveBaseAttachment

        pdf_bytes = self.body_to_pdf()
        filename = f"first-document-{self.thread_id}.pdf"
        existing = self.to_missiveattachment.filter(attachment_file__icontains=f"first-document-{self.thread_id}").first()
        if existing:
            existing.attachment_file.delete(save=False)
            existing.attachment_file.save(filename, ContentFile(pdf_bytes), save=True)
            return existing
        att = MissiveBaseAttachment.objects.create(
            missive=self,
            attachment_type=MissiveAttachmentType.ATTACHMENT,
            attachment_file=ContentFile(pdf_bytes, name=filename),
            priority=0,
            linked=False,
        )
        return att

    @property
    def subject_compiled(self):
        """Compile the subject of the missive."""
        context = self.missive_context()
        tpl = self.get_locally_or_campaign_value("subject")
        return Template(tpl).render(Context(context))

    @property
    def body_html_compiled(self):
        """Compile the HTML body of the missive."""
        context = self.missive_context()
        tpl = self.get_locally_or_campaign_value("body_html")
        return Template(tpl).render(Context(context))

    @property
    def body_text_compiled(self):
        """Compile the body text of the missive."""
        context = self.missive_context()
        tpl = self.get_locally_or_campaign_value("body_text")
        return Template(tpl).render(Context(context))

    @property
    def body_sms_compiled(self):
        """Compile the body SMS of the missive."""
        context = self.missive_context()
        tpl = self.get_locally_or_campaign_value("body_sms", self.body_text)
        return Template(tpl).render(Context(context))

    @property
    def first_document_compiled(self):
        """Compile first_document (campaign) or fallback to body_html when no campaign."""
        context = self.missive_context()
        tpl = self.get_locally_or_campaign_value("first_document", self.body_html) or ""
        return Template(str(tpl)).render(Context(context))

    #########################################################
    # Attachments
    #########################################################

    @property
    def base_url(self):
        """Base URL for attachments and other needs. Uses get_base_url() from settings."""
        return get_base_url(trailing_slash=False)

    @property
    def show_attachments_linked(self):
        """Show the attachments linked."""
        html = "<div>"
        for attachment in self.get_serialized_attachments(linked=True):
            data = {
                "url": attachment["url"],
                "icon": ATTACHMENT_ICON,
                "name": attachment["name"],
                "style": ATTACHMENT_STYLE,
            }
            html += ATTACHMENT_TPL_HTML.format(**data)
        html += "</div>"
        return mark_safe(html)  # nosec B703 B308

    @property
    def show_attachments_linked_text(self):
        """Show the attachments linked text."""
        qs = self.get_serialized_attachments(linked=True)
        if not qs:
            return ""
        title = _("Attachments:")
        text = f"{title}{SEPARATOR}"
        for attachment in qs:
            text += (
                f"- {attachment['name']}\n{self.base_url}{attachment['url']}{SEPARATOR}"
            )
        return text

    @property
    def attachments(self):
        """Attachments from missive and campaign (when campaign_id is set)."""
        from .attachment import MissiveBaseAttachment
        q_filter = models.Q(attachment_type=MissiveAttachmentType.ATTACHMENT) | models.Q(
            attachment_type=MissiveAttachmentType.VIRTUAL_ATTACHMENT
        )
        parent_q = models.Q(missive=self)
        if self.campaign_id:
            parent_q |= models.Q(campaign=self.campaign)
        return MissiveBaseAttachment.objects.filter(parent_q, q_filter)

    @property
    def attachments_physical(self):
        return self.attachments.filter(linked=False)

    def get_serialized_attachments(self, linked=False):
        """Get the attachments of the missive."""
        if not linked and self.missive_type == MissiveType.LRE:
            self.generate_first_document()
        return [a.get_serialized_attachment(linked=linked) for a in self.attachments.filter(linked=linked)]

    #########################################################
    # Services
    #########################################################

    @transaction.atomic
    def resend_missive(self):
        """Resend the missive: original becomes HISTORY, new duplicate is MISSIVE and gets sent."""
        if not self.can_resend():
            raise ValidationError(_("Missive cannot be resend"))
        self.thread_type = MissiveThreadType.HISTORY
        self.save(update_fields=["thread_type"])
        new_missive = self.duplicate_missive(thread_type=MissiveThreadType.MISSIVE, thread_id=self.thread_id, resend=True)
        new_missive.send_missive(old_missive=self)
        return new_missive

    def duplicate_attachments(self, new_missive, source_missive):
        """Copy attachments from source_missive to new_missive (excl. first-page)."""
        first_doc_prefix = f"first-document-{source_missive.thread_id}"
        attachments = source_missive.attachments.filter(
            attachment_type=MissiveAttachmentType.ATTACHMENT,
            missive=source_missive
        ).exclude(priority=0)
        for index, attachment in enumerate(attachments):
            attachment.pk = None
            attachment.id = None
            attachment.external_id = None
            attachment.missive = new_missive
            attachment.priority = index + 1
            attachment.save()

    def duplicate_recipients(self, new_missive, source_missive):
        """Copy recipients from source_missive to new_missive."""
        for recipient in source_missive.to_missiverecipient.all():
            recipient.pk = None
            recipient.id = None
            recipient.external_id = None
            recipient.missive = new_missive
            recipient.save()

    def duplicate_related_objects(self, new_missive, source_missive):
        """Copy related objects (e.g. contact links) from source_missive to new_missive."""
        for rel_obj in source_missive.to_missiverelatedobject.all():
            rel_obj.pk = None
            rel_obj.id = None
            rel_obj.missive = new_missive
            rel_obj.save()

    @transaction.atomic
    def duplicate_missive(self, thread_type=MissiveThreadType.MISSIVE, thread_id=None, resend=False):
        """Duplicate the missive with its attachments, recipients and related objects."""
        # Preserve source before mutating (new_missive = self would overwrite self)
        ModelClass = type(self)
        source = ModelClass.objects.get(pk=self.pk)
        missive_pre_duplicate.send(
            sender=ModelClass,
            source_missive=source,
            resend=resend,
            thread_type=thread_type,
            thread_id=thread_id,
        )
        new_missive = ModelClass.objects.get(pk=self.pk)
        new_missive.pk = None
        new_missive.id = None
        new_missive.external_id = None
        new_missive.thread_id = thread_id or uuid.uuid4()
        new_missive.thread_type = thread_type
        new_missive.status = MissiveStatus.DRAFT
        new_missive.save()
        self.duplicate_attachments(new_missive, source)
        self.duplicate_recipients(new_missive, source)
        self.duplicate_related_objects(new_missive, source)
        if resend:
            self.clear_campaign_sourced_fields(new_missive)
        missive_post_duplicate.send(
            sender=ModelClass,
            source_missive=source,
            new_missive=new_missive,
            resend=resend,
        )
        return new_missive

    def _update_recipients(self, recipients):
        for recipient in recipients:
            rec = self.to_missiverecipient.get(id=recipient.get("internal_id"))
            rec.external_id = recipient.get("external_id")
            rec.save(update_fields=["external_id"])

    def _update_attachments(self, attachments):
        for attachment in attachments:
            att = self.to_missiveattachment.get(id=attachment.get("internal_id"))
            att.external_id = attachment.get("external_id")
            att.save(update_fields=["external_id"])

    def prepare_missive(self):
        """Prepare the missive for sending (calls provider create)."""
        response = self.call_provider_service("create", **self.get_serialized_data())
        response["client_initiated"] = True
        self.external_id = response.get("external_id")
        self.save(update_fields=["external_id"])
        self._update_recipients(response.get("recipients", []))

    def update_missive(self):
        """Update the missive."""
        response = self.call_provider_service("update", **self.get_serialized_data())
        response["client_initiated"] = True
        self._update_recipients(response.get("recipients", []))

    def send_missive(self, *, old_missive=None):
        """Send the missive.

        :param old_missive: When sending a duplicate after a resend, pass the previous missive
            row (typically HISTORY). None for a normal first send.
        """
        if not self.can_send():
            raise ValidationError(_("Missive cannot be sent"))
        missive_pre_send.send(sender=self.__class__, missive=self, old_missive=old_missive)
        self.set_locally_ifnull()
        self.status = MissiveStatus.PROCESSING
        occurred_at = timezone.now()
        response = self.call_provider_service("send", **self.get_serialized_data())
        response["client_initiated"] = True
        if response.get("recipients"):
            self._update_recipients(response.get("recipients"))
        if response.get("attachments"):
            self._update_attachments(response.get("attachments"))
        self._update_attachments(response.get("attachments", []))
        self.external_id = response.get("external_id")
        if self.external_id:
            self.external_id = response.get("external_id")
            self.save(update_fields=["external_id", "status"])
            self.to_missiveevent.create(
                event=MissiveEventType.REQUEST,
                trace=response,
                client_initiated=True,
                occurred_at=occurred_at,
            )
            events = response.get("events")
            if events:
                self.handle_events(events)
        else:
            self.to_missiveevent.create(
                event=MissiveEventType.ERROR,
                trace=response,
                client_initiated=True,
                occurred_at=occurred_at,
            )
        self.refresh_from_db()
        missive_post_send.send(sender=self.__class__, missive=self, old_missive=old_missive)

    def handle_events(self, events: list | dict):
        from ..events import handle_events
        handle_events(events, self.provider, self.missive_type)

    def cancel_missive(self):
        """Cancel the missive."""
        response = self.call_provider_service("cancel", **self.get_serialized_data(attachments=False))
        if response.get("code") in [200, 204, 404]:
            self.status = MissiveStatus.CANCELLED
            self.save(update_fields=["status"])

    def retrieve_missive(self):
        """Retrieve the status of the missive from the provider."""
        response = self.call_provider_service("retrieve", **self.get_serialized_data(attachments=False))
        events = response.get("events")
        if events:
            self.handle_events(events)

    def set_status(self):
        from ..models.event import MissiveEvent

        success_count, processing_count, failed_count = MissiveEvent.objects.get_event_counts(missive=self)
        status = status_from_event_counts(success_count, processing_count, failed_count)
        if status != self.status:
            self.status = status
            self.save(update_fields=["status"])

    #########################################################
    # Billing
    #########################################################

    def can_billings(self):
        return self.has_service("get_billings") and self.external_id

    def get_billings(self):
        """Get the billings of the missive."""
        if self.can_billings():
            from ..billings import handle_billings
            handle_billings(**self.get_serialized_data(attachments=False))

    def set_billed(self):
        """Set the billed status on billing records for this missive."""
        self.to_missivebilling.filter(billing_amount__gt=0).update(is_billed=True)

    #########################################################
    # Proofs
    #########################################################

    def can_proofs(self):
        """Return True if provider supports get_proofs for this missive type."""
        return self

    def get_proofs(self):
        """Get proofs (filename, url) from provider. Returns [] if not supported."""
        if not self.can_proofs():
            return []
        provider = self.provider._provider
        service_name = f"retrieve_proofs_{self.missive_type}"
        if not hasattr(provider, service_name):
            return []
        return provider.call_service_formatted(
            service_name, **self.get_serialized_data(attachments=False)
        )

    def download_proof(self, **kwargs):
        """Download the proof from the provider."""
        if not self.can_proofs():
            return None
        provider = self.provider._provider
        service_name = f"download_proof_{self.missive_type}"
        if not hasattr(provider, service_name):
            return None
        return provider.call_service_formatted(service_name, output_format="raw", **kwargs)

    #########################################################
    # Recipients
    #########################################################

    @property
    def recipients(self):
        return self.to_missiverecipient.filter(
            recipient_type=MissiveRecipientType.RECIPIENT
        )

    @property
    def first_recipient(self):
        try:
            return self.recipients.first()
        except ObjectDoesNotExist:
            return _("Unknown recipient")

    @property
    def cc(self):
        return self.to_missiverecipient.filter(recipient_type=MissiveRecipientType.CC)

    @property
    def bcc(self):
        return self.to_missiverecipient.filter(recipient_type=MissiveRecipientType.BCC)

    #########################################################
    # Clean methods
    #########################################################

    def clean(self):
        """Clean the missive."""
        clean_by_support = f"clean_support_{self.missive_support}".lower()
        if hasattr(self, clean_by_support):
            getattr(self, clean_by_support)()

    def clean_subject(self):
        if not self.subject and not self.campaign:
            raise ValidationError({
                "subject": _("Subject or Campaign is required"),
            })

    def clean_support_email(self):
        """Clean the missive for email support."""
        has_body_missive = (self.body_html or self.body_text)
        has_body_campaign = (self.campaign and (self.campaign.body_html or self.campaign.body_text))
        if not has_body_missive and not has_body_campaign:
            raise ValidationError({
                "body_html": _("Body or body text is required (in missive or campaign)"),
                "body_text": _("Body or body text is required (in missive or campaign)"),
            })

    def clean_support_phone(self):
        """Clean the missive for SMS support."""
        has_body_missive = self.body_text
        has_body_campaign = (self.campaign and self.campaign.body_sms)
        if not has_body_missive and not has_body_campaign:
            raise ValidationError({
                "body_text": _("Body text is required (in missive or campaign)"),
            })

    def clean_support_address(self):
        """Clean the missive for address support."""
        has_body_missive = (self.body_html or self.to_missiveattachment.all().exists())
        has_body_campaign = (self.campaign and (self.campaign.first_document or self.campaign.to_campaigndocument.exists()))
        if not has_body_missive and not has_body_campaign:
            raise ValidationError({
                "body_html": _("Body or attachments are required (in missive or campaign)"),
            })


class MissiveHistory(Missive):
    """Missive history model."""
    objects = MissiveHistoryManager()

    class Meta:
        proxy = True
        verbose_name = _("Missive History")
        verbose_name_plural = _("Missive Histories")
        ordering = ["-created_at"]


class MissiveMessage(Missive):
    """Missive message model."""
    objects = MissiveMessageManager()

    class Meta:
        proxy = True
        verbose_name = _("Missive Message")
        verbose_name_plural = _("Missive Messages")
        ordering = ["-created_at"]


"""Billing handling: fetch via provider.get_billings_{missive_type}, then process each billing."""
from .models.provider import MissiveProviderModel
from .models.billing import MissiveBilling
from .models.missive import Missive
from .utils import get_recipient


def _process_billing(missive, bill):
    lookup = {
        "missive": missive,
        "invoice": bill.get("invoice"),
        "billing_amount": bill.get("billing_amount"),
        "estimate_amount": bill.get("estimate_amount"),
    }
    if bill.get("recipient"):
        lookup["recipient"] = get_recipient(missive, bill.get("recipient"))
    defaults = {
        "currency": bill.get("currency"),
        "trace": bill.get("raw") or {},
    }
    MissiveBilling.objects.update_or_create(defaults=defaults, **lookup)


def handle_billings(**kwargs) -> None:
    """Fetch billings from provider and process each one."""
    provider = kwargs.get('provider')
    provider = MissiveProviderModel.objects.get(name=provider)
    service_name = f"get_billings_{kwargs.get('missive_type')}"
    if not hasattr(provider._provider, service_name):
        return
    billings = provider._provider.call_service_formatted(
        service_name, **kwargs
    )
    if not billings:
        return
    external_id = kwargs.get('external_id')
    try:
        missive = Missive.objects.get(external_id=external_id)
    except Missive.DoesNotExist:
        return
    for bill in billings:
        _process_billing(missive, bill)



"""Admin for MissiveConfig."""

from django.contrib import admin

from ..models.config import MissiveConfig


@admin.register(MissiveConfig)
class MissiveConfigAdmin(admin.ModelAdmin):
    list_display = ["missive_type", "default_provider"]
    list_editable = ["default_provider"]
    ordering = ["missive_type"]



from typing import List

from django.urls import URLPattern, path

from .views.attachment import MissiveAttachmentDownloadView
from .views.preview import DownloadPDFView, PreviewFormView, PreviewView
from .views.webhook import WebhookView

app_name = "django_pymissive"

urlpatterns: List[URLPattern] = [
    path(
        "preview/<str:campaign_or_missive>/<uuid:pk>/",
        PreviewView.as_view(),
        name="preview",
    ),
    path(
        "preview/<str:campaign_or_missive>/",
        PreviewFormView.as_view(),
        name="preview_form",
    ),
    path(
        "webhook/<str:provider>/<str:missive_type>/",
        WebhookView.as_view(),
        name="missive_webhook",
    ),
    path(
        "attachment/<str:campaign_or_missive>/<uuid:pk>/download/",
        MissiveAttachmentDownloadView.as_view(),
        name="missive_attachment_download",
    ),
    path(
        "pdf/<str:campaign_or_missive>/<uuid:pk>/",
        DownloadPDFView.as_view(),
        name="download_pdf",
    ),
]



"""Webhook model for storing webhook configurations."""

from django.db import models
from django.utils.translation import gettext_lazy as _
from .choices import MissiveType, WebhookScheme

from pymissive.config import WEBHOOK_FIELDS
from ..utils import get_default_domain, get_default_scheme, build_webhook_url
from django_providerkit import fields_associations, ProviderField
from ..managers.webhook import MissiveWebhookManager


def get_default_webhook_scheme():
    """Return default scheme for webhook (serializable for migrations)."""
    return WebhookScheme.HTTPS if get_default_scheme() == "https" else WebhookScheme.HTTP

class MissiveWebhook(models.Model):
    """Webhook configuration for missive events."""

    provider = ProviderField(
        package_name="pymissive",
        blank=True,
        null=True,
        verbose_name=_("Provider"),
        help_text=_("Provider used to send this missive"),
    )
    webhook_id = models.CharField(
        max_length=255,
        verbose_name=_("Webhook ID"),
        help_text=_("Webhook ID"),
        primary_key=True,
    )
    scheme = models.CharField(
        max_length=5,
        choices=WebhookScheme.choices,
        default=get_default_webhook_scheme,
        blank=True,
        verbose_name=_("Scheme"),
        help_text=_("HTTP or HTTPS"),
    )
    domain = models.CharField(
        max_length=255,
        blank=True,
        default=get_default_domain,
        verbose_name=_("Domain"),
        help_text=_(
            "Base domain (e.g. example.com). Webhook path is built from provider and type."
        ),
    )

    objects = MissiveWebhookManager()

    class Meta:
        managed = False
        verbose_name = _("Webhook")
        verbose_name_plural = _("Webhooks")
        ordering = ["-created_at"]

    def __str__(self):
        return self.webhook_id

    def get_provider(self):
        if getattr(self.provider, "_provider", None):
            return self.provider
        from ..models.provider import MissiveProviderModel
        provider = self.provider_name
        return MissiveProviderModel.objects.get(name=provider)

    @property
    def provider_name(self):
        if getattr(self, "webhook_id", None):
            return self.webhook_id.split("-")[0]
        provider = getattr(self, "provider", None)
        return getattr(provider, "name", None) or str(provider) if provider else ""

    def _get_url(self):
        """Return URL from provider data, or build from scheme+domain."""
        scheme = getattr(self, "scheme", None) or "https"
        domain = getattr(self, "domain", None)
        if domain:
            base = f"{scheme}://{(domain or '').strip().lstrip('/')}"
            return build_webhook_url(base, self.provider_name, self.type)
        return ""

    def get_webhook_data(self):
        url = getattr(self, "url", None) or self._get_url()
        return {"id": self.id, "webhook_id": self.webhook_id, "type": self.type, "url": url}

    def new_webhook(self):
        service = f"create_webhook_{self.type}"
        provider = self.get_provider()
        if hasattr(provider._provider, service):
            return provider._provider.call_service(
                service, webhook_data=self.get_webhook_data()
            )

    def update_webhook(self):
        service = f"update_webhook_{self.type}"
        provider = self.get_provider()
        if hasattr(provider._provider, service):
            return provider._provider.call_service(
                service, webhook_data=self.get_webhook_data()
            )

    def save(self, *args, **kwargs):
        self.webhook_id = (
            self.new_webhook() if not self.webhook_id else self.update_webhook()
        )

    def delete(self):
        service = f"delete_webhook_{self.type}".lower()
        provider = self.get_provider()
        if hasattr(provider._provider, service):
            provider._provider.call_service(
                service, webhook_data=self.get_webhook_data()
            )


for field, cfg in WEBHOOK_FIELDS.items():
    if field not in ("webhook_id", "scheme", "domain"):
        field_cfg = {
            "verbose_name": cfg["label"],
            "help_text": cfg["description"],
        }
        if field == "type":
            field_cfg["choices"] = MissiveType.choices
        db_field = fields_associations[cfg["format"]](**field_cfg)
        MissiveWebhook.add_to_class(field, db_field)



"""Models for django_pymissive."""

from .config import MissiveConfig
from .billing import MissiveBilling
from .attachment import (
    MissiveBaseAttachment,
    MissiveAttachment,
    MissiveVirtualAttachment,
    CampaignAttachment,
    CampaignVirtualAttachment,
    MissiveProof,
)
from .campaign import MissiveCampaign, MissiveScheduledCampaign
from .choices import (
    AcknowledgementLevel,
    MissiveEventType,
    MissivePriority,
    MissiveStatus,
    MissiveType,
    WebhookScheme,
)
from .event import MissiveEvent
from .missive import Missive
from .provider import MissiveProviderModel
from .related_object import MissiveRelatedObject
from .webhook import MissiveWebhook
from .recipient import (
    MissiveRecipient,
    MissiveRecipientEmail,
    MissiveRecipientPhone,
    MissiveRecipientAddress,
    MissiveRecipientApplication,
)

__all__ = [
    "MissiveBilling",
    "MissiveConfig",
    "CampaignAttachment",
    "CampaignVirtualAttachment",
    "MissiveCampaign",
    "MissiveScheduledCampaign",
    "MissiveProviderModel",
    "Missive",
    "MissiveBaseAttachment",
    "MissiveAttachment",
    "MissiveVirtualAttachment",
    "MissiveEvent",
    "MissiveRelatedObject",
    "MissiveWebhook",
    "MissiveRecipient",
    "MissiveRecipientEmail",
    "MissiveRecipientPhone",
    "MissiveRecipientAddress",
    "MissiveRecipientApplication",
    "MissiveType",
    "MissiveEventType",
    "MissiveStatus",
    "MissivePriority",
    "AcknowledgementLevel",
    "WebhookScheme",
]



from django_providerkit.admin.service import ProviderServiceAdmin
from django_pymissive.models.service import MissiveServiceModel
from django.contrib import admin

@admin.register(MissiveServiceModel)
class MissiveServiceAdmin(ProviderServiceAdmin):
    """Admin for missive services."""
    pass


"""Missive campaign models."""

import uuid

from django.db import models, transaction
from django.template import Context, Template
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ..managers.campaign import MissiveCampaignManager
from ..models.base import CommentTimestampedModel
from ..models.choices import MissiveStatus, MissivePriority, AcknowledgementLevel, MissiveDeliveryMode
from django_geoaddress.fields import GeoaddressField
from phonenumber_field.modelfields import PhoneNumberField
from ..fields import RichTextField, JSONField
from ..utils import get_base_url


class MissiveCampaign(CommentTimestampedModel):
    """Campaign grouping missives for batch sending."""
    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        verbose_name=_("ID"),
    )
    subject = models.CharField(
        max_length=255,
        verbose_name=_("Subject"),
        help_text=_("Campaign subject"),
    )
    description = RichTextField(
        blank=True,
        verbose_name=_("Description"),
        help_text=_("Campaign description"),
    )

    # Email
    acknowledgement_email = models.CharField(
        max_length=50,
        choices=AcknowledgementLevel.choices,
        default=AcknowledgementLevel.BASIC_DELIVERY,
        verbose_name=_("Acknowledgement Level"),
        help_text=_("Desired acknowledgement level for delivery proof"),
    )
    sender_email_name = models.CharField(
        max_length=255,
        verbose_name=_("Sender email name"),
        help_text=_("Campaign sender email name"),
        blank=True,
        null=True,
    )
    sender_email = models.EmailField(
        verbose_name=_("Sender email"),
        help_text=_("Campaign sender email"),
        blank=True,
        null=True,
    )
    reply_to_email_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name=_("Reply-To name"),
        help_text=_("Display name for reply-to address"),
    )
    reply_to_email = models.EmailField(
        blank=True,
        null=True,
        verbose_name=_("Reply-To email"),
        help_text=_("Email address for replies"),
    )
    body_html = RichTextField(
        blank=True,
        verbose_name=_("Body HTML"),
        help_text=_("Campaign email HTML body"),
    )
    body_text = models.TextField(
        blank=True,
        verbose_name=_("Body text"),
        help_text=_("Campaign body text"),
    )

    # SMS
    sender_phone_name = models.CharField(
        max_length=255,
        verbose_name=_("Sender phone name"),
        help_text=_("Campaign sender phone name"),
        blank=True,
        null=True,
    )
    sender_phone = PhoneNumberField(
        blank=True,
        null=True,
        verbose_name=_("Sender phone"),
        help_text=_("Phone number of the sender (used for SMS)"),
    )
    body_sms = models.TextField(
        blank=True,
        verbose_name=_("Body SMS"),
        help_text=_("Campaign body SMS"),
    )

    # Address / LRE
    sender_address_name = models.CharField(
        max_length=255,
        verbose_name=_("Sender address name"),
        help_text=_("Campaign sender address name"),
        blank=True,
        null=True,
    )
    sender_address = GeoaddressField(
        verbose_name=_("Sender address"),
        help_text=_("Campaign sender address"),
        blank=True,
        null=True,
    )
    reply_to_address_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name=_("Reply-To address name"),
        help_text=_("Display name for reply-to address"),
    )
    reply_to_address = GeoaddressField(
        max_length=512,
        blank=True,
        null=True,
        verbose_name=_("Reply-To address"),
        help_text=_("Postal address for replies"),
    )
    acknowledgement_lre = models.CharField(
        max_length=50,
        choices=AcknowledgementLevel.choices,
        default=AcknowledgementLevel.BASIC_DELIVERY,
        verbose_name=_("Acknowledgement Level"),
        help_text=_("Desired acknowledgement level for delivery proof"),
    )
    delivery_mode_lre = models.CharField(
        max_length=50,
        choices=MissiveDeliveryMode.choices,
        default=MissiveDeliveryMode.NORMAL,
        verbose_name=_("Delivery Mode"),
        help_text=_("Delivery mode (economic, normal, premium, express)"),
    )
    priority_lre = models.CharField(
        max_length=20,
        choices=MissivePriority.choices,
        default=MissivePriority.NORMAL,
        verbose_name=_("Priority"),
        help_text=_("Priority level"),
    )
    first_document = RichTextField(
        blank=True,
        verbose_name=_("First Document"),
        help_text=_("First document content (HTML, converted to PDF for LRE)"),
    )

    additional_context = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Additional context"),
        help_text=_("Additional context as JSON"),
    )
    additional_config = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Additional configuration"),
        help_text=_("Additional configuration as JSON"),
    )
    metadata = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Metadata"),
        help_text=_("Additional metadata as JSON"),
    )

    objects = MissiveCampaignManager()
    # Plain manager for select_for_update (PostgreSQL rejects FOR UPDATE with GROUP BY)
    objects_plain = models.Manager()

    class Meta:
        verbose_name = _("Campaign")
        verbose_name_plural = _("Campaigns")
        ordering = ["-created_at", "subject"]

    def __str__(self):
        return self.subject

    def campaign_context(self):
        """Context for template rendering."""
        return {}

    @property
    def base_url(self):
        """Base URL for attachments and other needs. Uses get_base_url() from settings."""
        return get_base_url(trailing_slash=False)

    @property
    def email_reply_to(self):
        """Reply-to dict for email; None when no reply address."""
        if not self.reply_to_email:
            return None
        return {
            "name": self.reply_to_email_name or "",
            "email": str(self.reply_to_email),
        }

    @property
    def address_reply_to(self):
        return {
            "name": self.reply_to_address_name or "",
            "address": self.reply_to_address or "",
        }

    @property
    def phone_sender(self):
        return {
            "name": self.sender_phone_name or "",
            "phone": self.sender_phone or "",
        }

    @property
    def email_sender(self):
        return {
            "name": self.sender_email_name or "",
            "email": self.sender_email or "",
        }
    @property
    def address_sender(self):
        return {
            "name": self.sender_address_name or "",
            "address": self.sender_address or "",
        }

    def body_html_compiled(self):
        """Render body HTML (email) with campaign context."""
        if not self.body_html:
            return ""
        return Template(str(self.body_html)).render(Context(self.campaign_context()))

    def body_text_compiled(self):
        """Render body_text (email plain text) with campaign context."""
        if not self.body_text:
            return ""
        return Template(str(self.body_text)).render(Context(self.campaign_context()))

    def first_document_compiled(self):
        """Render first_document with campaign context."""
        if not self.first_document:
            return ""
        return Template(str(self.first_document)).render(Context(self.campaign_context()))

    @property
    def attachments(self):
        from .choices import MissiveAttachmentType
        from django.db.models import Q
        return self.to_campaigndocument.filter(
            Q(attachment_type=MissiveAttachmentType.ATTACHMENT)
            | Q(attachment_type=MissiveAttachmentType.VIRTUAL_ATTACHMENT),
        )

    @property
    def attachments_physical(self):
        return self.attachments.filter(linked=False)

    def get_serialized_attachments(self, linked=False):
        """Get the attachments of the campaign."""
        qs = self.attachments.filter(linked=linked)
        return [attachment.get_serialized_attachment(linked=linked) for attachment in qs]

    def start_campaign(self):
        """Start the campaign."""
        with transaction.atomic():
            campaign = MissiveCampaign.objects_plain.select_for_update().get(pk=self.pk)
            if campaign.metadata.get("processing"):
                from django.core.exceptions import ValidationError
                raise ValidationError(_("Campaign is already being processed."))
            campaign.metadata = {**dict(campaign.metadata), "processing": True}
            campaign.save(update_fields=["metadata"])
            scheduled = campaign.to_missivecampaignsend.create(
                campaign=campaign,
                scheduled_send_date=timezone.now()
            )
            scheduled.start_scheduled_campaign()


class MissiveScheduledCampaign(CommentTimestampedModel):
    """Scheduled send for a campaign."""

    campaign = models.ForeignKey(
        MissiveCampaign,
        on_delete=models.CASCADE,
        related_name="to_missivecampaignsend",
        verbose_name=_("Campaign"),
        editable=False,
    )
    scheduled_send_date = models.DateTimeField(
        null=True,
        blank=True,
        verbose_name=_("Scheduled send date"),
        help_text=_(
            "Scheduled send date for the campaign (leave blank for immediate sending)"
        ),
    )
    send_date = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("Send date"),
        help_text=_("Actual send date for the campaign"),
    )
    ended_at = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("Ended at"),
        help_text=_("Actual ended date for the campaign"),
    )
    additional_config = JSONField(
        default=dict,
        blank=True,
        verbose_name=_("Additional configuration"),
        help_text=_("Additional configuration as JSON"),
    )

    class Meta:
        verbose_name = _("Campaign send")
        verbose_name_plural = _("Campaign sends")
        ordering = ["-scheduled_send_date", "-ended_at", "-id"]

    def start_scheduled_campaign(self):
        """Start the scheduled campaign."""
        from ..task import get_campaign_backend
        backend = get_campaign_backend()
        backend.delay(self.id)

    def run_campaign(self):
        """Run the campaign."""
        missives = self.campaign.to_missive.filter(status=MissiveStatus.DRAFT)
        for missive in missives:
            missive.send_missive()



"""Admin for MissiveRelatedObject model."""

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from django_boosted import AdminBoostModel
from django.utils.safestring import mark_safe
from django.urls import reverse

from ..models.related_object import MissiveRelatedObject, CampaignRelatedObject


class BaseRelatedObjectAdmin:
    """Base admin for related objects."""

    def object_url_change(self, obj):
        """Return the url for the related object."""
        try:
            return reverse(
                f"admin:{obj.content_type.app_label}_{obj.content_type.model}_change",
                args=[obj.object_id],
            )
        except Exception:
            return None
    
    @admin.display(description=_("Object URL Link"))
    def object_url_link_display(self, obj):
        """Return a link to the related object in the admin site."""
        if not obj or not obj.content_type_id or not obj.object_id:
            return "-"
        label = obj.object_str or f"{obj.content_type} #{obj.object_id}"
        url = self.object_url_change(obj)
        if not url:
            return label
        return mark_safe(f'<a href="{url}">{label}</a>')


class MissiveRelatedObjectInline(admin.TabularInline, BaseRelatedObjectAdmin):
    """Inline for missive related objects."""

    model = MissiveRelatedObject
    extra = 0
    fields = [
        "content_type",
        "object_id",
        "object_url_link_display",
    ]
    readonly_fields = ["object_url_link_display",]


@admin.register(MissiveRelatedObject)
class MissiveRelatedObjectAdmin(AdminBoostModel, BaseRelatedObjectAdmin):
    """Admin for missive related object model."""

    list_display = [
        "missive",
        "content_type",
        "object_url_link_display",
        "created_at",
    ]
    list_filter = [
        "content_type",
        "created_at",
    ]
    search_fields = [
        "missive__subject",
        "missive__recipients__name",
        "missive__recipients__email",
        "missive__recipients__phone",
        "missive__recipients__address",
    ]
    readonly_fields = [
        "object_url_link_display",
        "created_at",
    ]
    raw_id_fields = ["missive",]

    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(
            None,
            ["missive", "content_type", "object_id", "object_url_link_display", "comment"],
        )
        self.add_to_fieldset(
            _("Timestamps"),
            ["created_at"],
        )


class CampaignRelatedObjectInline(admin.TabularInline, BaseRelatedObjectAdmin):
    """Inline for campaign related objects."""

    model = CampaignRelatedObject
    extra = 0
    fields = [
        "content_type",
        "object_id",
        "object_url_link_display",
    ]
    readonly_fields = ["object_url_link_display",]



@admin.register(CampaignRelatedObject)
class CampaignRelatedObjectAdmin(AdminBoostModel, BaseRelatedObjectAdmin):
    """Admin for campaign related object model."""

    list_display = [
        "campaign",
        "content_type",
        "object_url_link_display",
        "created_at",
    ]
    list_filter = [
        "content_type",
        "created_at",
    ]
    search_fields = [
        "campaign__name",
    ]
    readonly_fields = [
        "object_url_link_display",
        "created_at",
        "updated_at",
    ]
    raw_id_fields = ["campaign",]

    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(
            None,
            ["campaign", "content_type", "object_id", "object_url_link_display"],
        )
        self.add_to_fieldset(
            _("Comment/Timestamps"),
            ["comment", "created_at", "updated_at"],
        )


"""MissiveBilling model for billing records."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base import CommentTimestampedModel


class MissiveBilling(CommentTimestampedModel):
    """Billing record for a missive, optionally scoped to a recipient."""

    missive = models.ForeignKey(
        "django_pymissive.Missive",
        on_delete=models.CASCADE,
        related_name="to_missivebilling",
        verbose_name=_("Missive"),
        help_text=_("Missive associated with this billing record"),
    )

    recipient = models.ForeignKey(
        "django_pymissive.MissiveRecipient",
        on_delete=models.CASCADE,
        related_name="to_recipientbilling",
        verbose_name=_("Recipient"),
        help_text=_("Recipient when billing is per-recipient; null for missive-level billing"),
        blank=True,
        null=True,
    )

    billing_amount = models.DecimalField(
        max_digits=10,
        decimal_places=4,
        null=True,
        blank=True,
        verbose_name=_("Billing Amount"),
        help_text=_("Amount billed for this missive or recipient"),
    )

    estimate_amount = models.DecimalField(
        max_digits=10,
        decimal_places=4,
        null=True,
        blank=True,
        verbose_name=_("Estimate Amount"),
        help_text=_("Estimated amount before sending"),
    )

    is_billed = models.BooleanField(
        default=False,
        verbose_name=_("Billed"),
        help_text=_("Indicates if this record has been billed"),
    )

    currency = models.CharField(
        max_length=3,
        blank=True,
        null=True,
        verbose_name=_("Currency"),
        help_text=_("ISO 4217 currency code (e.g. EUR, USD)"),
    )

    invoice = models.TextField(
        blank=True,
        null=True,
        verbose_name=_("Invoice"),
        help_text=_("Invoice or billing details from the provider"),
    )

    trace = models.JSONField(
        blank=True,
        null=True,
        verbose_name=_("Trace"),
        help_text=_("Trace of the billing request"),
    )

    class Meta:
        verbose_name = _("Billing")
        verbose_name_plural = _("Billings")
        ordering = ["-created_at"]

    def __str__(self):
        if self.recipient_id:
            return f"{self.missive_id} / {self.recipient_id} - {self.billing_amount or 0}"
        return f"{self.missive_id} - {self.billing_amount or 0}"



"""Admin for MissiveEvent model."""

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from django_boosted import AdminBoostModel
from urllib.parse import unquote

from django.contrib import messages

from ..models.event import MissiveEvent

class UntreatedListFilter(admin.SimpleListFilter):
    """Custom filter for untreated events."""

    title = _("Untreated")
    parameter_name = "untreated"

    def lookups(self, request, model_admin):
        return [("1", _("Yes")), ("0", _("No"))]

    def queryset(self, request, queryset):
        if self.value() == "1":
            return queryset.filter(missive__isnull=True)
        if self.value() == "0":
            return queryset.filter(missive__isnull=False)
        return queryset


class MissiveEventInline(admin.TabularInline):
    """Inline for missive events (read-only)."""

    model = MissiveEvent
    extra = 0
    readonly_fields = [
        "missive",
        "recipient",
        "event",
        "reason",
        "occurred_at",
        "client_initiated",
    ]
    fields = [
        "missive",
        "recipient",
        "event",
        "reason",
        "occurred_at",
        "client_initiated",
    ]
    show_change_link = True
    can_delete = False

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(MissiveEvent)
class MissiveEventAdmin(AdminBoostModel):
    """Admin for missive event model."""

    list_display = [
        "event",
        "missive",
        "recipient",
        "missive__provider",
        "occurred_at",
        "client_initiated",
    ]
    list_filter = [
        "event",
        "missive__provider",
        "client_initiated",
        UntreatedListFilter,
    ]
    search_fields = [
        "event",
        "reason",
        "missive__subject",
        "recipient__name",
        "recipient__email",
        "recipient__phone",
        "recipient__address",
    ]
    readonly_fields = [
        "missive",
        "recipient",
        "event",
        "reason",
        "metadata",
        "trace",
        "occurred_at",
        "client_initiated",
        "created_at",
        "updated_at",
    ]
    raw_id_fields = ["missive", "recipient"]
    changeform_actions = {
        "replay": _("Replay"),
    }

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "missive",
                    "recipient",
                    "event",
                    "reason",
                )
            },
        ),
    ]

    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(
            _("Details"),
            ["occurred_at", "trace", "client_initiated", "metadata",],
        )
        self.add_to_fieldset(_("Comment/Timestamps"), ["comment", "created_at", "updated_at"])

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_replay_permission(self, request, obj=None):
        return obj and obj.pk and obj.can_replay()

    def handle_replay(self, request, obj=None):
        """Handle replay of event."""
        obj = unquote(obj)
        obj = self.get_object(request, obj)
        obj.replay()
        messages.success(request, _("Event replayed successfully."))



# Generated by Django 6.0.2 on 2026-03-18 11:53

import django_pymissive.models.attachment
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "django_pymissive",
            "0003_missivebilling_delete_missiverecipientnotification_and_more",
        ),
    ]

    operations = [
        migrations.CreateModel(
            name="MissiveProof",
            fields=[],
            options={
                "verbose_name": "Proof",
                "verbose_name_plural": "Proofs",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("django_pymissive.missivebaseattachment",),
        ),
        migrations.AlterModelOptions(
            name="missivecampaign",
            options={
                "ordering": ["-created_at", "subject"],
                "verbose_name": "Campaign",
                "verbose_name_plural": "Campaigns",
            },
        ),
        migrations.AlterField(
            model_name="missivebaseattachment",
            name="attachment_file",
            field=models.FileField(
                blank=True,
                help_text="Leave blank if the attachment is hosted externally",
                max_length=2000,
                null=True,
                storage=django_pymissive.models.attachment.ConfigurableAttachmentStorage(),
                upload_to=django_pymissive.models.attachment._attachment_upload_to,
                verbose_name="Attachment File",
            ),
        ),
        migrations.AlterField(
            model_name="missivebaseattachment",
            name="priority",
            field=models.PositiveIntegerField(
                default=1,
                help_text="Page order. 0=first-document (letter body), others use 1, 2, 3... (0 reserved)",
                verbose_name="Priority",
            ),
        ),
    ]



"""
Create a dataset of missives for testing/development.

Loads fixtures: fixture_email.json, fixture_phone.json, fixture_postal.json
- Email: 2 recipients, 2 CC, 2 BCC, 1 sender, 1 reply-to
- Phone (SMS): 2 recipients
- Postal: 2 recipients

Usage:
  ./manage.py create_missive_dataset
  ./manage.py create_missive_dataset -f /path/to/custom_fixture.json
  ./manage.py create_missive_dataset -f fixture1.json -f fixture2.json

Or load fixtures directly:
  ./manage.py loaddata fixture_email fixture_phone fixture_postal
"""

from django.core.management import call_command
from django.core.management.base import BaseCommand

DEFAULT_FIXTURES = ["fixture_email", "fixture_phone", "fixture_postal"]


class Command(BaseCommand):
    help = (
        "Load missive dataset fixtures "
        "(email with 2 CC/2 BCC/2 recipients, phone with 2 recipients, postal with 2 recipients)"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-f",
            "--fixture",
            action="append",
            dest="fixtures",
            default=None,
            help="Path to fixture file (can be repeated for multiple files)",
        )

    def handle(self, *args, **options):
        fixtures = options.get("fixtures") or DEFAULT_FIXTURES
        call_command(
            "loaddata",
            *fixtures,
            verbosity=options.get("verbosity", 1),
        )
        self.stdout.write(
            self.style.SUCCESS(f"Dataset loaded: {len(fixtures)} fixture(s)")
        )



from django.db import models

from pymissive.config import SUCCESSFUL_EVENTS, FAILED_EVENTS, INFO_EVENTS


class MissiveEventManager(models.Manager):
    """Manager for the MissiveEvent model."""

    def get_event_counts(self, missive=None, recipient=None):
        """Return (success_count, processing_count, failed_count) from last event per recipient."""
        qs = self.filter(event__isnull=False)
        if missive is not None:
            qs = qs.filter(missive=missive)
        if recipient is not None:
            qs = qs.filter(recipient=recipient)
        qs = qs.order_by("recipient_id", "-occurred_at")
        by_recipient = {}
        for ev in qs:
            key = ev.recipient_id
            if key not in by_recipient:
                by_recipient[key] = ev.event
        success_count = sum(1 for e in by_recipient.values() if e in SUCCESSFUL_EVENTS)
        processing_count = sum(1 for e in by_recipient.values() if e in INFO_EVENTS)
        failed_count = sum(1 for e in by_recipient.values() if e in FAILED_EVENTS)
        return success_count, processing_count, failed_count



"""Admin for MissiveAttachment model."""

from django.contrib import admin

from django.utils.translation import gettext_lazy as _
from django_boosted import AdminBoostModel

from ..models.attachment import (
    MissiveBaseAttachment,
    MissiveAttachment,
    MissiveVirtualAttachment,
    CampaignAttachment,
    CampaignVirtualAttachment,
    MissiveProof,
)


@admin.register(MissiveBaseAttachment)
class MissiveAttachmentAdmin(AdminBoostModel):
    """Admin for missive attachment model."""

    list_display = [
        "id",
        "attachment_type",
        "missive",
        "priority",
    ]
    list_filter = [
        "missive__missive_type",
    ]
    search_fields = [
        "missive__subject",
        "missive__recipient_name",
        "campaign__name",
    ]
    readonly_fields = [
        "attachment_object",
        "created_at",
        "updated_at",
    ]
    raw_id_fields = ["missive", "campaign",]

    fieldsets = [
        (
            None,
            {
                "fields": (
                    "campaign",
                    "missive",
                    "priority",
                    "attachment_file",
                    "external_id",
                    "page_count",
                    "linked",
                )
            },
        ),
    ]

    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(
            _("Attachment Object"),
            [
                "attachment_content_type",
                "attachment_object_id",
                "attachment_object_arguments",
                "attachment_object",
            ],
        )
        self.add_to_fieldset(_("Comment/Timestamps"), ["comment", "created_at", "updated_at"])
        self.add_to_fieldset(_("Configs"), ["metadata"])


class MissiveAttachmentInline(admin.TabularInline):
    """Inline for missive attachments."""

    model = MissiveAttachment
    extra = 0
    fields = [
        "priority",
        "attachment_file",
        "external_id",
        "page_count",
        "linked",
    ]


class MissiveVirtualAttachmentInline(admin.TabularInline):
    """Inline for missive virtual attachments."""

    model = MissiveVirtualAttachment
    extra = 0
    fields = [
        "priority",
        "attachment_content_type",
        "attachment_object_id",
        "attachment_object_arguments",
        "external_id",
        "page_count",
        "linked",
    ]


class CampaignAttachmentInline(admin.TabularInline):
    """Inline for campaign attachments."""

    model = CampaignAttachment
    extra = 0
    fields = [
        "priority",
        "attachment_file",
        "external_id",
        "page_count",
        "linked",
    ]


class CampaignVirtualAttachmentInline(admin.TabularInline):
    """Inline for campaign virtual attachments."""

    model = CampaignVirtualAttachment
    extra = 0
    fields = [
        "priority",
        "attachment_content_type",
        "attachment_object_id",
        "attachment_object_arguments",
        "external_id",
        "page_count",
        "linked",
    ]

class MissiveProofInline(admin.TabularInline):
    """Inline for missive proofs."""

    model = MissiveProof
    extra = 0
    fields = [
        "attachment_file",
    ]
    readonly_fields = [
        "attachment_file",
    ]



from django.db import models
from django.db.models.expressions import Subquery, OuterRef
from django.db.models import F, Max, Q, Sum
from django.db.models.functions import Coalesce

from ..models.choices import (
    MissiveThreadType,
    MissiveStatus,
)


class BaseMissiveManager(models.Manager):
    """Manager for the Missive model."""

    def last_event_subquery(self, field: str = "event"):
        from ..models.event import MissiveEvent

        return Subquery(
            MissiveEvent.objects.filter(
                missive=OuterRef("pk"),
            )
            .order_by("-occurred_at", "-id")
            .values(field)[:1],
            output_field=models.CharField(),
        )

    def last_scheduled_subquery(self, field: str = "event"):
        from ..models.campaign import MissiveScheduledCampaign

        return Subquery(
            MissiveScheduledCampaign.objects.filter(
                campaign=OuterRef("campaign_id"),
                send_date__isnull=False,
            )
            .order_by(f"-{field}", "-id")
            .values(field)[:1],
            output_field=models.CharField(),
        )

    def is_billable_expr(self):
        return models.Case(
            models.When(total_billing_amount__gt=0, then=True),
            default=False,
            output_field=models.BooleanField(),
        )

    def is_billed_expr(self):
        return models.Case(
            models.When(
                Q(total_billing_amount__gt=0)
                & Q(total_billing_amount=F("total_billed_amount")),
                then=True,
            ),
            default=False,
            output_field=models.BooleanField(),
        )

    def total_billing_expr(self, field: str, is_billed: bool = False):
        from ..models.billing import MissiveBilling
        q_billed = Q(is_billed=True, missive_id=OuterRef("id")) if is_billed else Q(missive_id=OuterRef("id"))
        return Subquery(
            MissiveBilling.objects.filter(q_billed)
            .order_by()
            .values("missive_id")
            .annotate(total=Sum(field))
            .values("total")[:1],
            output_field=models.DecimalField(max_digits=10, decimal_places=4),
        )

    def get_queryset_annotated(self):
        qs = super().get_queryset()
        qs = qs.select_related("campaign")
        qs = qs.prefetch_related(
            "to_missiverecipient",
            "to_missiveattachment",
            "to_missiveevent",
            "to_missivebilling",
            "to_missiverelatedobject",
        )
        qs = qs.annotate(
            last_campaign_send_date=self.last_scheduled_subquery("send_date"),
            last_campaign_ended_at=self.last_scheduled_subquery("ended_at"),
            count_recipient=models.Count("to_missiverecipient", distinct=True),
            count_event=models.Count("to_missiveevent", distinct=True),
            last_event=self.last_event_subquery(field="event"),
            last_event_reason=self.last_event_subquery(field="reason"),
            last_event_date=Coalesce(Max("to_missiveevent__occurred_at"), F("created_at")),
            count_related_object=models.Count("to_missiverelatedobject", distinct=True),
            count_attachment=models.Count("to_missiveattachment", distinct=True),
            total_billing_amount=self.total_billing_expr("billing_amount"),
            total_estimate_amount=self.total_billing_expr("estimate_amount"),
            total_billed_amount=self.total_billing_expr("billing_amount", is_billed=True),
            **{
                f"count_recipient_{status.value.lower()}": models.Count(
                    "to_missiverecipient",
                    distinct=True,
                    filter=Q(to_missiverecipient__status=status),
                )
                for status in MissiveStatus
            },
        )
        qs = qs.annotate(
            is_billable=self.is_billable_expr(),
            is_billed=self.is_billed_expr(),
        )
        return qs


class MissiveManager(BaseMissiveManager):
    """Manager for the Missive model."""


    def count_missive_thread(self, thread_type: MissiveThreadType):
        qs = self.model._base_manager.get_queryset().filter(
            thread_type=thread_type, thread_id=OuterRef("thread_id")
        )
        return Subquery(
            qs.values("thread_id").annotate(count=models.Count("id")).values("count"),
            output_field=models.IntegerField(),
        )

    def get_queryset(self):
        qs = super().get_queryset_annotated()
        qs = qs.annotate(
            count_history=self.count_missive_thread(thread_type=MissiveThreadType.HISTORY),
            count_message=self.count_missive_thread(thread_type=MissiveThreadType.MESSAGE),
        )
        return qs


class MissiveHistoryManager(BaseMissiveManager):
    """Manager for the MissiveHistory model."""

    def get_queryset(self):
        qs = super().get_queryset_annotated()
        qs = qs.filter(thread_type=MissiveThreadType.HISTORY)
        return qs


class MissiveMessageManager(BaseMissiveManager):
    """Manager for the MissiveMessage model."""

    def get_queryset(self):
        qs = super().get_queryset_annotated()
        qs = qs.select_related("message_by")
        qs = qs.filter(thread_type=MissiveThreadType.MESSAGE)
        return qs


"""Admin configuration for django_pymissive."""

from .config import MissiveConfigAdmin
from .billing import MissiveBillingAdmin, MissiveBillingInline
from .attachment import MissiveAttachmentAdmin
from .campaign import MissiveCampaignAdmin
from .event import MissiveEventAdmin
from .recipient import (
    MissiveRecipientAdmin,
    MissiveRecipientEmailInline,
    MissiveRecipientPhoneInline,
    MissiveRecipientAddressInline,
    MissiveRecipientApplicationInline
)
from .missive import MissiveAdmin
from .provider import ProviderAdmin
from .related_object import MissiveRelatedObjectAdmin
from .webhook import MissiveWebhookAdmin
from .service import MissiveServiceAdmin

__all__ = [
    "MissiveBillingAdmin",
    "MissiveBillingInline",
    "MissiveConfigAdmin",
    "ProviderAdmin",
    "MissiveCampaignAdmin",
    "MissiveAdmin",
    "MissiveAttachmentAdmin",
    "MessageAdmin",
    "MessageInline",
    "MissiveEventAdmin",
    "MissiveRelatedObjectAdmin",
    "MissiveRecipientAdmin",
    "MissiveRecipientEmailInline",
    "MissiveRecipientPhoneInline",
    "MissiveRecipientAddressInline",
    "MissiveRecipientApplicationInline",
    "MissiveWebhookAdmin",
    "MissiveServiceAdmin",
]



"""Admin for MissiveRecipient model."""

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from django_boosted import AdminBoostModel

from ..models.recipient import (
    MissiveRecipient,
    MissiveRecipientEmail,
    MissiveRecipientPhone,
    MissiveRecipientAddress,
    MissiveRecipientApplication,
)
from ..models.choices import get_missive_style
from django.utils.html import format_html


class MissiveRecipientInline(admin.TabularInline):
    """Inline for missive recipients."""

    model = MissiveRecipient
    extra = 0
    readonly_fields = ["sent_at", "delivered_at"]
    fields = [
        "recipient_type",
        "status",
        "name",
        "email",
        "phone",
        "address",
        "notification_id",
        "external_id",
        "sent_at",
        "delivered_at",
    ]

class MissiveRecipientEmailInline(admin.TabularInline):
    """Inline for missive recipient emails."""

    model = MissiveRecipientEmail
    extra = 0
    readonly_fields = ["created_at", "updated_at", "sent_at", "delivered_at"]
    fields = [
        "recipient_type",
        "status",
        "name",
        "email",
        "external_id",
        "sent_at",
        "delivered_at",
    ]

class MissiveRecipientPhoneInline(admin.TabularInline):
    """Inline for missive recipient phones."""

    model = MissiveRecipientPhone
    extra = 0
    readonly_fields = ["created_at", "updated_at", "sent_at", "delivered_at"]
    fields = [
        "recipient_type",
        "status",
        "name",
        "phone",
        "external_id",
        "sent_at",
        "delivered_at",
    ]

class MissiveRecipientAddressInline(admin.TabularInline):
    """Inline for missive recipient addresses."""

    model = MissiveRecipientAddress
    extra = 0
    readonly_fields = ["created_at", "updated_at", "sent_at", "delivered_at"]
    fields = [
        "recipient_type",
        "status",
        "name",
        "address",
        "external_id",
        "sent_at",
        "delivered_at",
    ]

class MissiveRecipientApplicationInline(admin.TabularInline):
    """Inline for missive recipient applications (push, branded)."""

    model = MissiveRecipientApplication
    extra = 0
    readonly_fields = ["created_at", "updated_at", "sent_at", "delivered_at"]
    fields = [
        "recipient_type",
        "status",
        "name",
        "notification_id",
        "external_id",
        "sent_at",
        "delivered_at",
    ]


@admin.register(MissiveRecipient)
class MissiveRecipientAdmin(AdminBoostModel):
    """Admin for missive recipients."""

    list_display = [
        "recipient_display",
        "recipient_support",
        "missive_display",
        "recipient_type_display",
    ]
    list_filter = [
        "recipient_support",
        "recipient_type",
        "status",
    ]
    search_fields = [
        "missive__subject",
        "name",
        "email",
        "phone",
        "address",
    ]
    readonly_fields = [
        "created_at",
        "updated_at",
        "status",
        "sent_at",
        "delivered_at",
    ]
    raw_id_fields = [
        "missive",
    ]

    def change_fieldsets(self):
        """Configure fieldsets for change view."""
        self.add_to_fieldset(None, ["missive", "recipient_support", "recipient_type", "status", "name"])
        self.add_to_fieldset(_("Target"), ["email", "phone", "address", "notification_id", "external_id"])
        self.add_to_fieldset(_("Comment/Timestamps"), ["comment", "created_at", "updated_at"])

    def recipient_display(self, obj):
        """Display the recipient name and email or phone or address."""
        help_text = obj.email or obj.phone or obj.address
        return self.format_with_help_text(obj.name, help_text)

    recipient_display.short_description = _("Recipient")

    def recipient_type_display(self, obj):
        """Display the recipient type."""
        recipient_type_style = get_missive_style(obj.recipient_type)
        recipient_type = self.format_label(
            obj.get_recipient_type_display(),
            size="small",
            label_type=recipient_type_style,
        )
        status_style = get_missive_style(obj.status)
        status = self.format_label(
            obj.get_status_display(), size="small", label_type=status_style
        )
        html = format_html("{} {}", recipient_type, status)
        if obj.last_event:
            event_style = get_missive_style(obj.last_event)
            event = self.format_label(
                obj.last_event_reason, size="small", label_type=event_style
            )
            html = format_html("{} {}", html, event)
        return self.format_with_help_text(html, obj.last_event_date)

    recipient_type_display.short_description = _("Recipient Type")

    def missive_display(self, obj):
        """Display the missive subject."""
        return self.format_with_help_text(
            obj.missive.subject, obj.missive.get_missive_type_display()
        )

    missive_display.short_description = _("Missive")



"""Event handling: normalize via provider.handle_webhook_{missive_type}, then process each event."""

from datetime import timezone as dt_timezone

from django.conf import settings
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from .models.choices import MissiveEventType
from .models.event import MissiveEvent
from .models.missive import Missive
from .utils import get_recipient


def _can_save_untreated(provider_name):
    CONF_SAVE_UNTREATED = getattr(settings, "PYMISSIVE_SAVE_UNTREATED_EVENTS", False)
    if CONF_SAVE_UNTREATED is True:
        return True
    if isinstance(CONF_SAVE_UNTREATED, list):
        return provider_name in CONF_SAVE_UNTREATED
    return False


def _get_occurred_at(occurred_at):
    if isinstance(occurred_at, str):
        occurred_at = parse_datetime(occurred_at.replace("Z", "+00:00"))
    if occurred_at is not None and timezone.is_naive(occurred_at):
        return timezone.make_aware(occurred_at, dt_timezone.utc)
    return occurred_at or timezone.now()


def _save_untreated(event, provider):
    provider_name = getattr(provider, "name", None) or str(provider)
    if not _can_save_untreated(provider_name):
        return
    trace = {"event": event, "provider": provider_name}
    MissiveEvent.objects.create(
        missive=None,
        recipient=None,
        event=MissiveEventType.ERROR,
        reason=event.get("reason", "Could not process event"),
        occurred_at=_get_occurred_at(event.get("occurred_at")),
        trace=trace,
    )


def _process_event(event, missive):
    occurred_at = _get_occurred_at(event.get("occurred_at"))
    lookup = {
        "missive": missive,
        "event": event.get("event"),
        "occurred_at": occurred_at,
    }
    defaults = {
        "reason": event.get("reason", "No reason provided"),
        "trace": event.get("raw") or {},
    }
    recipient = None
    if event.get("recipient"):
        recipient = get_recipient(missive, event.get("recipient"))
        lookup["recipient"] = recipient

    raw = event.get("raw") or {}
    if "pk" in raw:
        defaults = {
            **defaults,
            **lookup,
        }
        lookup = {"pk": raw.get("pk")}
    MissiveEvent.objects.update_or_create(defaults=defaults, **lookup)
    if recipient:
        recipient.set_status()
    missive.set_status()


def handle_event(event, provider, missive_type: str) -> Missive | None:
    try:
        external_id = event.get("external_id")
        missive = Missive.objects.get(external_id=external_id)
        _process_event(event, missive)
    except Missive.DoesNotExist:
        _save_untreated(event, provider)
    return None


def handle_events(events, provider, missive_type: str) -> Missive | None:
    """Normalize via provider.handle_webhook_{missive_type}, then process each event."""
    events_normalized = provider._provider.call_service_formatted(
        f"handle_webhook_{missive_type}", payload=events
    )
    if events_normalized:
        if isinstance(events_normalized, dict):
            events_normalized = [events_normalized]
        for event in events_normalized:
            handle_event(event, provider, missive_type)
    return None
